# -*- coding: utf-8 -*-
from .Utils import getWordTag, readDictionary


def computeAccuracy(goldStandardCorpus, taggedCorpus):
    tagged = open(taggedCorpus, "r", encoding="utf-8-sig").read().split()
    goldStandard = open(goldStandardCorpus, "r", encoding="utf-8-sig").read().split()
    if len(tagged) != len(goldStandard):
        msg = f"The numbers of word tokens in {goldStandardCorpus} and {taggedCorpus} are not equal!"
        return 0, msg

    numwords = 0
    count = 0
    for i in range(len(tagged)):
        numwords += 1
        word1, tag1 = getWordTag(tagged[i])
        word2, tag2 = getWordTag(goldStandard[i])
        if word1 != word2 and word1 != "''" and word2 != "''":
            msg = f"Words are not the same in gold standard and tagged corpora, at the index {i}"
            return 0, msg

        if tag1.lower() == tag2.lower():
            count += 1

    return count * 100.0 / numwords


def computeAccuracies(fullDictFile, goldStandardCorpus, taggedCorpus):
    """
    Return known-word accuracy, unknown-word accuracy and the overall accuracy  
    """
    tagged = open(taggedCorpus, "r", encoding="utf-8-sig").read().split()
    goldStandard = open(goldStandardCorpus, "r", encoding="utf-8-sig").read().split()
    if len(tagged) != len(goldStandard):
        return f"The numbers of word tokens in {goldStandardCorpus} and {taggedCorpus} are not equal!"

    fullDICT = readDictionary(fullDictFile)

    numwords = count = 0
    countKN = countUNKN = 0
    countCorrectKN = countCorrectUNKN = 0

    for i in range(len(tagged)):
        numwords += 1
        word1, tag1 = getWordTag(tagged[i])
        word2, tag2 = getWordTag(goldStandard[i])
        if word1 != word2 and word1 != "''" and word2 != "''":
            return (
                "Words are not the same in gold standard and tagged corpora, at the index "
                + str(i)
            )

        if tag1.lower() == tag2.lower():
            count += 1

        if word1 in fullDICT:
            countKN += 1
            if tag1.lower() == tag2.lower():
                countCorrectKN += 1
        else:
            countUNKN += 1
            if tag1.lower() == tag2.lower():
                countCorrectUNKN += 1

    if countUNKN == 0:
        return countCorrectKN * 100.0 / countKN, 0.0, count * 100.0 / numwords
    else:
        return (
            countCorrectKN * 100.0 / countKN,
            countCorrectUNKN * 100.0 / countUNKN,
            count * 100.0 / numwords,
        )


def evaluate(goldStandardCorpus, taggedCorpus, fullDictFile=None):
    if fullDictFile:
        return computeAccuracies(fullDictFile, goldStandardCorpus, taggedCorpus)
    else:
        return computeAccuracy(goldStandardCorpus, taggedCorpus)



# -*- coding: utf-8 -*-

from multiprocessing import Pool

from ..InitialTagger.InitialTagger4Bo import initializeCorpus, initializeSentence
from ..SCRDRlearner.Object import FWObject
from ..SCRDRlearner.SCRDRTree import SCRDRTree
from ..SCRDRlearner.SCRDRTreeLearner import SCRDRTreeLearner
from ..Utility.Config import NUMBER_OF_PROCESSES, THRESHOLD
from ..Utility.Utils import getWordTag, getRawText, readDictionary
from ..Utility.LexiconCreator import createLexicon


def unwrap_self_RDRPOSTagger(arg, **kwarg):
    return RDRPOSTagger.tagRawSentence(*arg, **kwarg)


class RDRPOSTagger(SCRDRTree):
    """
    RDRPOSTagger for a particular language
    """

    def __init__(self):
        super().__init__()
        self.root = None

    def tagRawSentence(self, DICT, rawLine):
        """

        :param DICT:
        :param rawLine:
        :return:
        """
        line = initializeSentence(DICT, rawLine)
        sen = []
        wordTags = line.split()
        for i in range(len(wordTags)):
            fwObject = FWObject.getFWObject(wordTags, i)
            word, tag = getWordTag(wordTags[i])
            node = self.findFiredNode(fwObject)
            if node.depth > 0:
                sen.append(word + "/" + node.conclusion)
            else:  # Fired at root, return initialized tag
                sen.append(word + "/" + tag)
        return " ".join(sen)

    def tagRawCorpus(self, DICT, rawCorpusPath):
        lines = open(rawCorpusPath, "r", encoding="utf-8-sig").readlines()
        # Change the value of NUMBER_OF_PROCESSES to obtain faster tagging process!
        pool = Pool(processes=NUMBER_OF_PROCESSES)
        taggedLines = pool.map(
            unwrap_self_RDRPOSTagger,
            zip([self] * len(lines), [DICT] * len(lines), lines),
        )
        outW = open(rawCorpusPath + ".TAGGED", "w", encoding="utf-8-sig")
        for line in taggedLines:
            outW.write(line + "\n")
        outW.close()
        return "Output file: " + rawCorpusPath + ".TAGGED"


def rdr(to_process, mode=None, model=None, lexicon=None, string=False, verbose=False):
    """
    :param to_process: either a string(!!!"tag" only!!!) or a filename. if string, be sure to set arg "string" to True
    :param mode: either train or tag
    :param model: !!!"tag" only!!! .RDR file to be used as model to POS tag to_process
    :param lexicon: !!!"tag" only!!! .DICT file to be used as model to POS tag to_process
    :param string: !!!"tag" only!!! considers "to_process" as a string to be tagged if True, considers it a filename otherwise.
    :param verbose: return the messages from training and tagging if True, not otherwise.
    """
    if not mode or mode not in ["train", "tag"]:
        raise SyntaxError("mode should either be train or tag.")

    if mode == "train":
        log = []
        log.append("\n====== Start ======")
        log.append(
            "\nGenerate from the gold standard training corpus a lexicon "
            + to_process
            + ".DICT"
        )
        log.append(createLexicon(to_process, "full"))
        log.append(createLexicon(to_process, "short"))
        log.append(
            "\nExtract from the gold standard training corpus a raw text corpus "
            + to_process
            + ".RAW"
        )
        getRawText(to_process, to_process + ".RAW")
        log.append(
            "\nPerform initially POS tagging on the raw text corpus, to generate "
            + to_process
            + ".INIT"
        )
        DICT = readDictionary(to_process + ".sDict")
        initializeCorpus(DICT, to_process + ".RAW", to_process + ".INIT")
        log.append(
            "\nLearn a tree model of rules for POS tagging from %s and %s"
            % (to_process, to_process + ".INIT")
        )
        rdrTree = SCRDRTreeLearner(THRESHOLD[0], THRESHOLD[1])
        log.extend(rdrTree.learnRDRTree(to_process + ".INIT", to_process, verbose=verbose))
        log.append("\nWrite the learned tree model to file " + to_process + ".RDR")
        rdrTree.writeToFile(to_process + ".RDR")
        log.append("\nDone!")
        return "".join(log) if verbose else 0

    if mode == "tag":
        log = []
        r = RDRPOSTagger()
        log.append("\n=> Read a POS tagging model from " + model)
        r.constructSCRDRtreeFromRDRfile(model)
        log.append("\n=> Read a lexicon from " + lexicon)
        DICT = readDictionary(lexicon)
        log.append("\n=> Perform POS tagging on " + to_process)
        if string:
            r.tagRawSentence(DICT, to_process)
        else:
            r.tagRawCorpus(DICT, to_process)
        return "".join(log) if verbose else 0



# -*- coding: utf-8 -*-

from .Node import Node
from .Object import FWObject


class SCRDRTree:
    """
    Single Classification Ripple Down Rules tree for Part-of-Speech and morphological tagging
    """

    def __init__(self, root=None):
        self.root = root

    def findDepthNode(self, node, depth):
        while node.depth != depth:
            node = node.father
        return node

    def classify(self, object):
        self.root.check(object)

    def writeToFileWithSeenCases(self, outFile):
        out = open(outFile, "w", encoding="utf-8-sig")
        self.root.writeToFileWithSeenCases(out, 0)
        out.close()

    def writeToFile(self, outFile):
        out = open(outFile, "w", encoding="utf-8-sig")
        self.root.writeToFile(out, 0)
        out.close()

    # Build tree from file containing rules using FWObject
    def constructSCRDRtreeFromRDRfile(self, rulesFilePath):

        self.root = Node(FWObject(False), "NN", None, None, None, [], 0)
        currentNode = self.root
        currentDepth = 0

        rulesFile = open(rulesFilePath, "r", encoding="utf-8-sig")
        lines = rulesFile.readlines()

        for i in range(1, len(lines)):
            line = lines[i]
            depth = 0
            for c in line:
                if c == "\t":
                    depth = depth + 1
                else:
                    break

            line = line.strip()
            if len(line) == 0:
                continue

            temp = line.find("cc")
            if temp == 0:
                continue

            condition = getCondition(line.split(" : ", 1)[0].strip())
            conclusion = getConcreteValue(line.split(" : ", 1)[1].strip())

            node = Node(condition, conclusion, None, None, None, [], depth)

            if depth > currentDepth:
                currentNode.exceptChild = node
            elif depth == currentDepth:
                currentNode.elseChild = node
            else:
                while currentNode.depth != depth:
                    currentNode = currentNode.father
                currentNode.elseChild = node

            node.father = currentNode
            currentNode = node
            currentDepth = depth

    def findFiredNode(self, fwObject):
        currentNode = self.root
        firedNode = None
        obContext = fwObject.context
        while True:
            # Check whether object satisfying the current node's condition
            cnContext = currentNode.condition.context
            notNoneIds = currentNode.condition.notNoneIds
            satisfied = True
            for i in notNoneIds:
                if cnContext[i] != obContext[i]:
                    satisfied = False
                    break

            if satisfied:
                firedNode = currentNode
                exChild = currentNode.exceptChild
                if exChild is None:
                    break
                else:
                    currentNode = exChild
            else:
                elChild = currentNode.elseChild
                if elChild is None:
                    break
                else:
                    currentNode = elChild
        return firedNode


#    def findFiredNodeInDepth(self, fwObject, depth):
#        currentNode = self.root
#        firedNode = None
#        while True:
#            if(currentNode.condition.isSatisfied(fwObject)):
#                firedNode = currentNode
#                if currentNode.exceptChild is None:
#                    break
#                else:
#                    currentNode = currentNode.exceptChild
#            else:
#                if currentNode.elseChild is None:
#                    break
#                else:
#                    currentNode = currentNode.elseChild
#            if currentNode.depth > depth:
#                break
#        return firedNode
#
#    #Count number of nodes in exception-structure levels
#    def countNodes(self, inDepth):
#        currentNode = self.root
#        nodeQueue = []
#        nodeQueue.append(currentNode)
#        count = 0
#        while len(nodeQueue) > 0:
#            currentNode = nodeQueue[0]
#            #Current node's depth is smaller than a given threshold
#            if currentNode.depth <= inDepth:
#                count += 1
#            if currentNode.exceptChild is not None:
#                nodeQueue.append(currentNode.exceptChild)
#            if currentNode.elseChild is not None:
#                nodeQueue.append(currentNode.elseChild)
#            nodeQueue = nodeQueue[1:]
#        return count


def getConcreteValue(str):
    if str.find('""') > 0:
        if str.find("Word") > 0:
            return "<W>"
        elif str.find("suffixL") > 0:
            return "<SFX>"
        else:
            return "<T>"
    return str[str.find('"') + 1 : len(str) - 1]


def getCondition(strCondition):
    condition = FWObject(False)
    for rule in strCondition.split(" and "):
        rule = rule.strip()
        key = rule[rule.find(".") + 1 : rule.find(" ")]
        value = getConcreteValue(rule)

        if key == "prevWord2":
            condition.context[0] = value
        elif key == "prevTag2":
            condition.context[1] = value
        elif key == "prevWord1":
            condition.context[2] = value
        elif key == "prevTag1":
            condition.context[3] = value
        elif key == "word":
            condition.context[4] = value
        elif key == "tag":
            condition.context[5] = value
        elif key == "nextWord1":
            condition.context[6] = value
        elif key == "nextTag1":
            condition.context[7] = value
        elif key == "nextWord2":
            condition.context[8] = value
        elif key == "nextTag2":
            condition.context[9] = value
        elif key == "suffixL1":
            condition.context[10] = value
        elif key == "suffixL2":
            condition.context[11] = value
        elif key == "suffixL3":
            condition.context[12] = value
    for i in range(13):
        if condition.context[i] is not None:
            condition.notNoneIds.append(i)
    return condition



# -*- coding: utf-8 -*-


class Node:
    """
    A class to represent the nodes in SCRDR tree
    """

    def __init__(
        self,
        condition,
        conclusion,
        father=None,
        exceptChild=None,
        elseChild=None,
        cornerstoneCases=[],
        depth=0,
    ):
        self.condition = condition
        self.conclusion = conclusion
        self.exceptChild = exceptChild
        self.elseChild = elseChild
        self.cornerstoneCases = cornerstoneCases
        self.father = father
        self.depth = depth

    def satisfied(self, object):
        return eval(self.condition)

    def executeConclusion(self, object):
        exec(self.conclusion)

    def appendCornerstoneCase(self, object):
        self.cornerstoneCases.append(object)

    def check(self, object):
        if self.satisfied(object):
            self.executeConclusion(object)
            if self.exceptChild != None:
                self.exceptChild.check(object)
        else:
            if self.elseChild != None:
                self.elseChild.check(object)

    def checkDepth(self, object, length):
        if self.depth <= length:
            if self.satisfied(object):
                self.executeConclusion(object)
                if self.exceptChild != None:
                    self.exceptChild.checkDepth(object, length)
            else:
                if self.elseChild != None:
                    self.elseChild.checkDepth(object, length)

    def findRealFather(self):
        node = self
        fatherNode = node.father
        while True and fatherNode != None:
            if fatherNode.exceptChild == node:
                break
            node = fatherNode
            fatherNode = node.father
        return fatherNode

    def addElseChild(self, node):
        fatherNode = self.findRealFather()
        for object in fatherNode.cornerstoneCases:
            if node.satisfied(object):
                print("The new rule fires the cornerstone cases of its father node!!!")
                self.findRealFather().cornerstoneCases.remove(object)
        self.elseChild = node
        return True

    def addExceptChild(self, node):
        for object in self.cornerstoneCases:
            if node.satisfied(object):
                print("The new rule fires the cornerstone cases of its father node!!!")
                self.cornerstoneCases.remove(object)
        self.exceptChild = node
        return True

    def writeToFileWithSeenCases(self, out, depth):
        space = tabStr(depth)
        out.write(space + self.condition + " : " + self.conclusion + "\n")
        for case in self.cornerstoneCases:
            out.write(" " + space + "cc: " + case.toStr() + "\n")
        if self.exceptChild != None:
            self.exceptChild.writeToFile(out, depth + 1)
        if self.elseChild != None:
            self.elseChild.writeToFile(out, depth)

    def writeToFile(self, out, depth):
        space = tabStr(depth)
        out.write(space + self.condition + " : " + self.conclusion + "\n")
        if self.exceptChild != None:
            self.exceptChild.writeToFile(out, depth + 1)
        if self.elseChild != None:
            self.elseChild.writeToFile(out, depth)


def tabStr(length):
    return "".join(["\t"] * length)



# -*- coding: utf-8 -*-


class Object:
    attributes = [
        "word",
        "tag",
        "prevWord2",
        "prevWord1",
        "nextWord1",
        "nextWord2",
        "prevTag2",
        "prevTag1",
        "nextTag1",
        "nextTag2",
        "suffixL1",
        "suffixL2",
        "suffixL3",
    ]
    code = "def __init__(self"
    for att in attributes:
        code = code + ", " + att + " = None"
    code = code + "):\n"
    for att in attributes:
        code = code + "    self." + att + "=" + att + "\n"

    exec(code)

    def toStr(self):
        res = "("
        for att in Object.attributes:
            boo = eval("isinstance(self. " + att + ", str)")
            if not boo:
                res = res + str(eval("self." + att))
            else:
                res = res + '"' + str(eval("self." + att)) + '"'

            if att != Object.attributes[len(Object.attributes) - 1]:
                res = res + ","
        res += ")"
        return res


def getWordTag(wordTag):
    if wordTag == "///":
        return "/", "/"
    index = wordTag.rfind("/")
    if index == -1:
        return None, None
    word = wordTag[:index].strip()
    tag = wordTag[index + 1 :].strip()
    return word, tag


def getObject(wordTags, index):  # Sequence of "Word/Tag"
    word, tag = getWordTag(wordTags[index])
    preWord1 = preTag1 = preWord2 = preTag2 = ""
    nextWord1 = nextTag1 = nextWord2 = nextTag2 = ""
    suffixL1 = suffixL2 = suffixL3 = ""

    decodedW = word
    syls = decodedW.split("་")
    if len(syls) >= 2:
        suffixL1 = "་".join(syls[-1:])
    if len(syls) >= 3:
        suffixL2 = "་".join(syls[-2:])
    if len(syls) >= 4:
        suffixL3 = "་".join(syls[-3:])

    if index > 0:
        preWord1, preTag1 = getWordTag(wordTags[index - 1])
    if index > 1:
        preWord2, preTag2 = getWordTag(wordTags[index - 2])
    if index < len(wordTags) - 1:
        nextWord1, nextTag1 = getWordTag(wordTags[index + 1])
    if index < len(wordTags) - 2:
        nextWord2, nextTag2 = getWordTag(wordTags[index + 2])

    return Object(
        word,
        tag,
        preWord2,
        preWord1,
        nextWord1,
        nextWord2,
        preTag2,
        preTag1,
        nextTag1,
        nextTag2,
        suffixL1,
        suffixL2,
        suffixL3,
    )


def getObjectDictionary(initializedCorpus, goldStandardCorpus):
    goldStandardSens = open(goldStandardCorpus, "r", encoding="utf-8-sig").readlines()
    initializedSens = open(initializedCorpus, "r", encoding="utf-8-sig").readlines()

    objects = {}

    j = 0
    for i in range(len(initializedSens)):
        init = initializedSens[i].strip()
        if len(init) == 0:
            continue

        while j < len(goldStandardSens) and goldStandardSens[j].strip() == "":
            j += 1

        if j >= len(goldStandardSens):
            continue

        gold = goldStandardSens[j].strip()
        j += 1

        initWordTags = (
            init.replace("“", "''").replace("”", "''").replace('"', "''").split()
        )
        goldWordTags = (
            gold.replace("“", "''").replace("”", "''").replace('"', "''").split()
        )

        for k in range(len(initWordTags)):
            initWord, initTag = getWordTag(initWordTags[k])
            goldWord, correctTag = getWordTag(goldWordTags[k])

            if initWord != goldWord:
                msg = "\nERROR (Raw texts are mismatched || Some sentence is incorrectly formatted):"
                msg += "\n" + str(i + 1) + "th initialized sentence:   " + " ".join(initWordTags)
                msg += "\n" + str(i + 1) + "th gold standard sentence: " + " ".join(goldWordTags)
                raise SyntaxError(msg)

            if initTag not in objects.keys():
                objects[initTag] = {}
                objects[initTag][initTag] = []

            if correctTag not in objects[initTag].keys():
                objects[initTag][correctTag] = []

            objects[initTag][correctTag].append(getObject(initWordTags, k))

    return objects


class FWObject:
    """
    RDRPOSTaggerV1.1: new implementation scheme
    RDRPOSTaggerV1.2: add suffixes
    """

    def __init__(self, check=False):
        self.context = [
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        ]
        if check == True:
            i = 0
            while i < 10:
                self.context[i] = "<W>"
                self.context[i + 1] = "<T>"
                i = i + 2
            self.context[10] = "<SFX>"  # suffix
            self.context[11] = "<SFX>"
            self.context[12] = "<SFX>"
        self.notNoneIds = []

    @staticmethod
    def getFWObject(startWordTags, index):
        object = FWObject(True)
        word, tag = getWordTag(startWordTags[index])
        object.context[4] = word
        object.context[5] = tag

        decodedW = word
        if len(decodedW) >= 4:
            object.context[10] = decodedW[-2:]
            object.context[11] = decodedW[-3:]
        if len(decodedW) >= 5:
            object.context[12] = decodedW[-4:]

        if index > 0:
            preWord1, preTag1 = getWordTag(startWordTags[index - 1])
            object.context[2] = preWord1
            object.context[3] = preTag1

        if index > 1:
            preWord2, preTag2 = getWordTag(startWordTags[index - 2])
            object.context[0] = preWord2
            object.context[1] = preTag2

        if index < len(startWordTags) - 1:
            nextWord1, nextTag1 = getWordTag(startWordTags[index + 1])
            object.context[6] = nextWord1
            object.context[7] = nextTag1

        if index < len(startWordTags) - 2:
            nextWord2, nextTag2 = getWordTag(startWordTags[index + 2])
            object.context[8] = nextWord2
            object.context[9] = nextTag2

        return object


#    def isSatisfied(self, fwObject):
#        for i in range(13):
#            key = self.context[i]
#            if (key is not None):
#                if key != fwObject.context[i]:
#                    return False
#        return True



# -*- coding: utf-8 -*-


def getWordTag(wordTag):
    if wordTag == "///":
        return "/", "/"
    index = wordTag.rfind("/")
    if index == -1:
        return wordTag, None
    word = wordTag[:index].strip()
    tag = wordTag[index + 1 :].strip()
    return word, tag


def getRawText(inputFile, outFile):
    out = open(outFile, "w", encoding="utf-8-sig")
    sents = open(inputFile, "r", encoding="utf-8-sig").readlines()
    for sent in sents:
        wordTags = sent.strip().split()
        for wordTag in wordTags:
            word, tag = getWordTag(wordTag)
            out.write(word + " ")
        out.write("\n")
    out.close()


def readDictionary(inputFile):
    dictionary = {}
    lines = open(inputFile, "r", encoding="utf-8-sig").readlines()
    for line in lines:
        wordtag = line.strip().split()
        dictionary[wordtag[0]] = wordtag[1]
    return dictionary



# -*- coding: utf-8 -*-

from .Node import Node
from .Object import getObjectDictionary
from .SCRDRTree import SCRDRTree

# Generate concrete rules based on input object of 5-word window context object


def generateRules(object):
    # 1. Current word
    rule1 = 'object.word == "' + object.word + '"'
    # 2. Next 1st word
    rule2 = 'object.nextWord1 == "' + object.nextWord1 + '"'
    # 3. Next 2nd word
    rule3 = 'object.nextWord2 == "' + object.nextWord2 + '"'
    # 4. Previous 1st word
    rule4 = 'object.prevWord1 == "' + object.prevWord1 + '"'
    # 5. Previous 2nd word
    rule5 = 'object.prevWord2 == "' + object.prevWord2 + '"'

    # 6. Current word and next 1st word
    rule6 = rule1 + " and " + rule2
    # 7. Previous 1st word and current word
    rule7 = rule4 + " and " + rule1
    # 11. Previous 1st word and next 1st word
    rule11 = rule4 + " and " + rule2
    # 29. Next 1st word and next 2nd word
    # rule29 = rule2 + " and " + rule3
    # 30. Previous 2nd word and previous 1st word
    # rule30 = rule5 + " and " + rule4
    # 19. Current word and next 2nd word
    rule19 = rule1 + " and " + rule3
    # 20. Previous 2nd word and current word
    rule20 = rule5 + " and " + rule1

    # 8. Current word, next 1st word and next 2nd word
    rule8 = rule6 + " and " + rule3
    # 9. Previous 2nd word, previous 1st word and current word
    rule9 = rule5 + " and " + rule7
    # 10. Previous 1st word, current word and next 1st word
    rule10 = rule4 + " and " + rule6

    # 12. Next 1st tag
    rule12 = 'object.nextTag1 == "' + object.nextTag1 + '"'
    # 13. Next 2nd tag
    rule13 = 'object.nextTag2 == "' + object.nextTag2 + '"'
    # 14. Previous 1st tag
    rule14 = 'object.prevTag1 == "' + object.prevTag1 + '"'
    # 15. Previous 2nd tag
    rule15 = 'object.prevTag2 == "' + object.prevTag2 + '"'
    # 16. Next 1st tag and next 2nd tag
    rule16 = rule12 + " and " + rule13
    # 17. Previous 2nd tag and previous 1st tag
    rule17 = rule15 + " and " + rule14
    # 18. Previous 1st tag and next 1st tag
    rule18 = rule14 + " and " + rule12

    # 21. Current word and next 1st tag
    rule21 = rule1 + " and " + rule12
    # 22. Current word and previous 1st tag
    rule22 = rule14 + " and " + rule1
    # 23. Previous 1st tag, current word and next 1st tag
    rule23 = rule14 + " and " + rule21
    # 24. Current word and 2 next tags.
    rule24 = rule1 + " and " + rule16
    # 25. 2 previous tags and current word
    rule25 = rule17 + " and " + rule1
    # 26. 1-syl suffix
    rule26 = 'object.suffixL1 == "' + object.suffixL1 + '"'
    # 27. 2-syl suffix
    rule27 = 'object.suffixL2 == "' + object.suffixL2 + '"'
    # 28. 3-syl suffix
    rule28 = 'object.suffixL3 == "' + object.suffixL3 + '"'

    rules = []
    rules.append(rule1)
    rules.append(rule2)
    rules.append(rule3)
    rules.append(rule4)
    rules.append(rule5)
    rules.append(rule6)
    rules.append(rule7)
    rules.append(rule8)
    rules.append(rule9)
    rules.append(rule10)
    rules.append(rule11)
    rules.append(rule12)
    rules.append(rule13)
    rules.append(rule14)
    rules.append(rule15)
    rules.append(rule16)
    rules.append(rule17)
    rules.append(rule18)
    rules.append(rule19)
    rules.append(rule20)
    rules.append(rule21)
    rules.append(rule22)
    rules.append(rule23)
    rules.append(rule24)
    rules.append(rule25)
    rules.append(rule26)
    rules.append(rule27)
    rules.append(rule28)
    # rules.append(rule29)
    # rules.append(rule30)

    rules = set(rules)
    return rules


def countMatching(objects, ruleNotIn):
    counts = {}
    matchedObjects = {}
    for object in objects:
        rules = generateRules(object)
        for rule in rules:
            if rule in ruleNotIn:
                continue
            counts[rule] = counts.setdefault(rule, 0) + 1
            matchedObjects.setdefault(rule, []).append(object)
    return counts, matchedObjects


def satisfy(object, rule):
    return eval(rule)


def fire(rule, cornerstoneCases):
    for object in cornerstoneCases:
        if satisfy(object, rule):
            return True
    return False


def generateRulesFromObjectSet(objects):
    res = []
    for object in objects:
        rules = generateRules(object)
        res += rules
    return res


class SCRDRTreeLearner(SCRDRTree):
    def __init__(self, iThreshold=2, mThreshold=2):
        super().__init__()
        self.improvedThreshold = iThreshold
        self.matchedThreshold = mThreshold

    # For layer-2 exception structure
    def findMostImprovingRuleForTag(
        self, startTag, correctTag, correctCounts, wrongObjects
    ):
        impCounts, affectedObjects = countMatching(wrongObjects, [])

        maxImp = -1000000
        bestRule = ""
        for rule in impCounts:
            temp = impCounts[rule]
            if rule in correctCounts:
                temp -= correctCounts[rule]

            if temp > maxImp:
                maxImp = temp
                bestRule = rule

        if maxImp == -1000000:
            affectedObjects[bestRule] = []

        return bestRule, maxImp, affectedObjects[bestRule]

    def findMostEfficientRule(self, startTag, objects, correctCounts):
        maxImp = -1000000
        rule = ""
        correctTag = ""
        cornerstoneCases = []

        for tag in objects:
            if tag == startTag:
                continue
            if (
                len(objects[tag]) <= maxImp
                or len(objects[tag]) < self.improvedThreshold
            ):
                continue

            ruleTemp, imp, affectedObjects = self.findMostImprovingRuleForTag(
                startTag, correctTag, correctCounts, objects[tag]
            )
            if imp >= self.improvedThreshold and imp > maxImp:
                maxImp = imp
                rule = ruleTemp
                correctTag = tag
                cornerstoneCases = affectedObjects

        needToCorrectObjects = {}
        errorRaisingObjects = []
        if maxImp > -1000000:
            for tag in objects:
                if tag != correctTag:
                    for object in objects[tag]:
                        if satisfy(object, rule):
                            needToCorrectObjects.setdefault(tag, []).append(object)
                            if tag == startTag:
                                errorRaisingObjects.append(object)

        return (
            rule,
            correctTag,
            maxImp,
            cornerstoneCases,
            needToCorrectObjects,
            errorRaisingObjects,
        )

    def findMostMatchingRule(self, matchingCounts):
        correctTag = ""
        bestRule = ""
        maxCount = -1000000

        for tag in matchingCounts:
            for rule in matchingCounts[tag]:
                if (
                    matchingCounts[tag][rule] >= self.matchedThreshold
                    and matchingCounts[tag][rule] > maxCount
                ):
                    maxCount = matchingCounts[tag][rule]
                    bestRule = rule
                    correctTag = tag

        return bestRule, correctTag

    def buildNodeForObjectSet(self, objects, root):
        cornerstoneCaseRules = generateRulesFromObjectSet(root.cornerstoneCases)

        matchingCounts = {}
        matchingObjects = {}
        for tag in objects:
            matchingCounts[tag], matchingObjects[tag] = countMatching(
                objects[tag], cornerstoneCaseRules
            )

        total = 0
        for tag in objects:
            total += len(objects[tag])

        currentNode = root
        elseChild = False
        while True:
            rule, correctTag = self.findMostMatchingRule(matchingCounts)

            if rule == "":
                break

            cornerstoneCases = matchingObjects[correctTag][rule]

            needToCorrectObjects = {}
            for tag in objects:
                if rule in matchingObjects[tag]:
                    if tag != correctTag:
                        needToCorrectObjects[tag] = matchingObjects[tag][rule]
                    for object in matchingObjects[tag][rule]:
                        rules = generateRules(object)
                        for rule1 in rules:
                            if rule1 not in matchingCounts[tag]:
                                continue
                            matchingCounts[tag][rule1] -= 1

            node = Node(
                rule,
                'object.conclusion = "' + correctTag + '"',
                currentNode,
                None,
                None,
                cornerstoneCases,
            )

            if not elseChild:
                currentNode.exceptChild = node
                elseChild = True
            else:
                currentNode.elseChild = node

            currentNode = node
            self.buildNodeForObjectSet(needToCorrectObjects, currentNode)

    def learnRDRTree(self, initializedCorpus, goldStandardCorpus, verbose):
        self.root = Node("True", 'object.conclusion = "NN"', None, None, None, [], 0)

        objects = getObjectDictionary(initializedCorpus, goldStandardCorpus)

        currentNode = self.root
        log = []
        for initializedTag in objects:
            if verbose:
                log.append(f"\n===> Building exception rules for tag {initializedTag}")
            correctCounts = {}
            for object in objects[initializedTag][initializedTag]:
                rules = generateRules(object)
                for rule in rules:
                    correctCounts[rule] = correctCounts.setdefault(rule, 0) + 1

            node = Node(
                'object.tag == "' + initializedTag + '"',
                'object.conclusion = "' + initializedTag + '"',
                self.root,
                None,
                None,
                [],
                1,
            )

            if self.root.exceptChild == None:
                self.root.exceptChild = node
            else:
                currentNode.elseChild = node

            currentNode = node
            objectSet = objects[initializedTag]

            elseChild = False
            currentNode1 = currentNode
            while True:
                rule, correctTag, imp, cornerstoneCases, needToCorrectObjects, errorRaisingObjects = self.findMostEfficientRule(
                    initializedTag, objectSet, correctCounts
                )
                if imp < self.improvedThreshold:
                    break

                node = Node(
                    rule,
                    'object.conclusion = "' + correctTag + '"',
                    currentNode,
                    None,
                    None,
                    cornerstoneCases,
                    2,
                )

                if not elseChild:
                    currentNode1.exceptChild = node
                    elseChild = True
                else:
                    currentNode1.elseChild = node

                currentNode1 = node

                for object in cornerstoneCases:
                    objectSet[correctTag].remove(object)

                for tag in needToCorrectObjects:
                    for object in needToCorrectObjects[tag]:
                        objectSet[tag].remove(object)

                for object in errorRaisingObjects:
                    rules = generateRules(object)
                    for rule in rules:
                        correctCounts[rule] -= 1

                self.buildNodeForObjectSet(needToCorrectObjects, currentNode1)

        return log



# -*- coding: utf-8 -*-

import re


def initializeSentence(FREQDICT, sentence):
    words = sentence.strip().split()
    taggedSen = []
    for word in words:
        if word in ["“", "”", '"']:
            # taggedSen.append("''/" + FREQDICT["''"])
            if "''" in FREQDICT:
                taggedSen.append("''/" + FREQDICT["''"])
            elif "." in FREQDICT:
                taggedSen.append("''/" + FREQDICT["."])
            elif "," in FREQDICT:
                taggedSen.append("''/" + FREQDICT[","])
            else:
                print(
                    "\n'' is not in the dictionary \nManually add '' with a possible POS tag into the .DICT file!"
                )
                taggedSen.append("''/" + FREQDICT["''"])
            continue

        tag = ""
        decodedW = word
        lowerW = decodedW.lower()
        if word in FREQDICT:
            tag = FREQDICT[word]
        elif lowerW in FREQDICT:
            tag = FREQDICT[lowerW]
        else:
            if re.search(r"[0-9]+", word) != None:
                tag = FREQDICT["TAG4UNKN-NUM"]
            else:
                suffixL2 = suffixL3 = suffixL4 = suffixL1 = None
                syls = decodedW.split("་")
                wLength = len(syls)
                if wLength >= 2:
                    suffixL1 = ".*" + "་".join(syls[-1:])
                if wLength >= 3:
                    suffixL2 = ".*" + "་".join(syls[-2:])
                if wLength >= 4:
                    suffixL3 = ".*" + "་".join(syls[-3:])
                if wLength >= 5:
                    suffixL4 = ".*" + "་".join(syls[-4:])

                if suffixL4 in FREQDICT:
                    tag = FREQDICT[suffixL4]
                elif suffixL3 in FREQDICT:
                    tag = FREQDICT[suffixL3]
                elif suffixL2 in FREQDICT:
                    tag = FREQDICT[suffixL2]
                elif suffixL1 in FREQDICT:
                    tag = FREQDICT[suffixL1]
                else:
                    tag = FREQDICT["TAG4UNKN-WORD"]

                # suffixL2 = suffixL3 = suffixL4 = suffixL5 = None
                # wLength = len(decodedW)
                # if wLength >= 4:
                #     suffixL3 = ".*" + decodedW[-3:]
                #     suffixL2 = ".*" + decodedW[-2:]
                # if wLength >= 5:
                #     suffixL4 = ".*" + decodedW[-4:]
                # if wLength >= 6:
                #     suffixL5 = ".*" + decodedW[-5:]

                # if suffixL5 in FREQDICT:
                #     tag = FREQDICT[suffixL5]
                # elif suffixL4 in FREQDICT:
                #     tag = FREQDICT[suffixL4]
                # elif suffixL3 in FREQDICT:
                #     tag = FREQDICT[suffixL3]
                # elif suffixL2 in FREQDICT:
                #     tag = FREQDICT[suffixL2]
                # elif decodedW[0].isupper():
                #     tag = FREQDICT["TAG4UNKN-CAPITAL"]
                # else:
                #     tag = FREQDICT["TAG4UNKN-WORD"]

        taggedSen.append(word + "/" + tag)

    return " ".join(taggedSen)


def initializeCorpus(FREQDICT, inputFile, outputFile):
    lines = open(inputFile, "r", encoding="utf-8-sig").readlines()
    fileOut = open(outputFile, "w", encoding="utf-8-sig")
    for line in lines:
        fileOut.write(initializeSentence(FREQDICT, line) + "\n")
    fileOut.close()



#! /usr/bin/env python
# -*- coding: utf8 -*-

from __future__ import print_function

from pathlib import Path
import setuptools
from pkg_resources import parse_version
import re

assert parse_version(setuptools.__version__) >= parse_version("38.6.0")


def get_version(prop, project):
    project = Path(__file__).parent / project / "__init__.py"
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), project.read_text()
    )
    return result.group(1)


def read(fname):
    p = Path(__file__).parent / fname
    with p.open(encoding="utf-8") as f:
        return f.read()


setuptools.setup(
    name="bordr",
    version=get_version("__version__", "bordr"),  # edit version in pybo/__init__.py
    author="Dat Quoc Nguyen",
    author_email="dqnguyen@unimelb.edu.au",
    description="A  fast and accurate POS and morphological tagging toolkit, lightly adapted to Tibetan language.",
    license="GNU General Public License",
    keywords="part-of-speech-tagger java nlp pos-tagging pos-tagger python3",
    url="https://github.com/Esukhia/RDRPOSTagger",
    packages=setuptools.find_packages(),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    project_urls={
        "Source": "https://github.com/Esukhia/RDRPOSTagger",
        "Tracker": "https://github.com/Esukhia/RDRPOSTagger/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Text Processing :: Linguistic",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: Tibetan",
    ],
    package_data={},
    python_requires=">=3.6",
    dependency_links=[],
    install_requires=[],
)



# -*- coding: utf-8 -*-
import re
from .Utils import getWordTag


def add2WordTagFreqDict(word, tag, inDict):
    if word not in inDict:
        inDict[word] = {}
        inDict[word][tag] = 1
    else:
        if tag not in inDict[word]:
            inDict[word][tag] = 1
        else:
            inDict[word][tag] += 1


def createLexicon(corpusFilePath, fullLexicon):
    if fullLexicon not in ["full", "short"]:
        raise SyntaxError(
            "The second parameter gets 'full' or 'short' string-value!\nNo lexicon is generated!!!"
        )
    # elif fullLexicon == 'full':
    #    print "Creating a full .DICT lexicon from the gold standard training corpus", corpusFilePath
    # else:
    #    print "Creating a short .sDict lexicon which excludes word types appearing 1 time in the gold standard training corpus"

    lines = open(corpusFilePath, "r", encoding="utf-8-sig").readlines()
    log = []
    wordTagCounter = {}
    for i in range(len(lines)):
        # print i
        pairs = (
            lines[i]
            .strip()
            .replace("“", "''")
            .replace("”", "''")
            .replace('"', "''")
            .split()
        )
        for pair in pairs:
            word, tag = getWordTag(pair)
            if (len(word) >= (len(pair) - 1)) or (len(tag) >= (len(pair) - 1)):
                if fullLexicon == "full":
                    log.append(
                        f"ERROR: The {i + 1}th sentence is incorrectly formatted!"
                    )
            else:
                add2WordTagFreqDict(word, tag, wordTagCounter)

    from operator import itemgetter

    dictionary = {}
    suffixDictCounter = {}

    tagCounter_Alphabet = {}
    tagCounter_CapitalizedWord = {}
    tagCounter_Numeric = {}

    for word in wordTagCounter:
        tagFreq4Word = wordTagCounter[word]
        pairs = list(tagFreq4Word.items())
        pairs.sort(key=itemgetter(1), reverse=True)
        tag = pairs[0][0]

        decodedWord = word
        isCapital = decodedWord[0].isupper()

        if fullLexicon == "full":
            dictionary[word] = tag
        else:  # Get the lexicon without 1-time-occurrence word types
            if (len(pairs) == 1 and pairs[0][1] > 1) or len(pairs) > 1:
                dictionary[word] = tag

        if re.search(r"[0-9]+", word) != None:
            if tag not in tagCounter_Numeric:
                tagCounter_Numeric[tag] = 1
            else:
                tagCounter_Numeric[tag] += 1
        else:
            if isCapital:
                if tag not in tagCounter_CapitalizedWord:
                    tagCounter_CapitalizedWord[tag] = 1
                else:
                    tagCounter_CapitalizedWord[tag] += 1
            else:
                if tag not in tagCounter_Alphabet:
                    tagCounter_Alphabet[tag] = 1
                else:
                    tagCounter_Alphabet[tag] += 1

            if len(decodedWord) >= 4:
                suffix = ".*" + decodedWord[-3:]
                add2WordTagFreqDict(suffix, tag, suffixDictCounter)
                suffix = ".*" + decodedWord[-2:]
                add2WordTagFreqDict(suffix, tag, suffixDictCounter)
            if len(decodedWord) >= 5:
                suffix = ".*" + decodedWord[-4:]
                add2WordTagFreqDict(suffix, tag, suffixDictCounter)
            if len(decodedWord) >= 6:
                suffix = ".*" + decodedWord[-5:]
                add2WordTagFreqDict(suffix, tag, suffixDictCounter)

    from collections import OrderedDict

    dictionary = OrderedDict(sorted(dictionary.items(), key=itemgetter(0)))

    # Get the most frequent tag in the lexicon to label unknown words and numbers
    tagCounter_Alphabet = OrderedDict(
        sorted(tagCounter_Alphabet.items(), key=itemgetter(1), reverse=True)
    )
    tagCounter_CapitalizedWord = OrderedDict(
        sorted(tagCounter_CapitalizedWord.items(), key=itemgetter(1), reverse=True)
    )
    tagCounter_Numeric = OrderedDict(
        sorted(tagCounter_Numeric.items(), key=itemgetter(1), reverse=True)
    )
    tag4UnknWord = list(tagCounter_Alphabet.keys())[0]
    tag4UnknCapitalizedWord = tag4UnknWord
    tag4UnknNum = tag4UnknWord
    if len(tagCounter_CapitalizedWord) > 0:
        tag4UnknCapitalizedWord = list(tagCounter_CapitalizedWord.keys())[0]
    if len(tagCounter_Numeric) > 0:
        tag4UnknNum = list(tagCounter_Numeric.keys())[0]

    # Write to file
    fileSuffix = ".sDict"
    if fullLexicon == "full":
        fileSuffix = ".DICT"
    fileOut = open(corpusFilePath + fileSuffix, "w", encoding="utf-8")

    fileOut.write("TAG4UNKN-WORD " + tag4UnknWord + "\n")
    fileOut.write("TAG4UNKN-CAPITAL " + tag4UnknCapitalizedWord + "\n")
    fileOut.write("TAG4UNKN-NUM " + tag4UnknNum + "\n")
    for key in dictionary:
        fileOut.write(key + " " + dictionary[key] + "\n")

    for suffix in suffixDictCounter:
        tagFreq4Suffix = suffixDictCounter[suffix]
        pairs = list(tagFreq4Suffix.items())
        pairs.sort(key=itemgetter(1), reverse=True)
        tag = pairs[0][0]
        freq = pairs[0][1]
        if len(suffix) == 7 and freq >= 2:
            fileOut.write(suffix + " " + tag + "\n")
        if len(suffix) == 6 and freq >= 3:
            fileOut.write(suffix + " " + tag + "\n")
        if len(suffix) == 5 and freq >= 4:
            fileOut.write(suffix + " " + tag + "\n")
        if len(suffix) == 4 and freq >= 5:
            fileOut.write(suffix + " " + tag + "\n")

    fileOut.close()
    return "".join(log)
